@@ -1,32 +1,41 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
-import uuid
+import datetime as dt
 import logging
 import urllib.parse
+import uuid
 from decimal import Decimal
 from email.header import Header
 from itertools import groupby
 from operator import attrgetter
 
 import stripe
+from sql import Literal
 
 from trytond.cache import Cache
 from trytond.config import config
 from trytond.i18n import gettext
 from trytond.model import (
-    ModelSQL, ModelView, Workflow, DeactivableMixin, fields, dualmethod)
-from trytond.pool import PoolMeta, Pool
-from trytond.pyson import Eval, Bool
+    DeactivableMixin, Index, ModelSQL, ModelView, Unique, Workflow, dualmethod,
+    fields)
+from trytond.modules.account_payment.exceptions import (
+    PaymentValidationError, ProcessError)
+from trytond.modules.company.model import (
+    employee_field, reset_employee, set_employee)
+from trytond.modules.currency.fields import Monetary
+from trytond.pool import Pool, PoolMeta
+from trytond.pyson import Eval, TimeDelta
 from trytond.report import Report, get_email
 from trytond.rpc import RPC
 from trytond.sendmail import sendmail_transactional
+from trytond.tools import sql_pairing
+from trytond.tools.email_ import set_from_header
 from trytond.transaction import Transaction
 from trytond.url import http_host
 from trytond.wizard import (
-    Wizard, StateAction, StateView, StateTransition, Button)
+    Button, StateAction, StateTransition, StateView, Wizard)
 
-from trytond.modules.account_payment.exceptions import (
-    ProcessError, PaymentValidationError)
+from .common import StripeCustomerMethodMixin
 
 logger = logging.getLogger(__name__)
 stripe.max_network_retries = config.getint(
@@ -44,8 +53,7 @@ class Journal(metaclass=PoolMeta):
         states={
             'required': Eval('process_method') == 'stripe',
             'invisible': Eval('process_method') != 'stripe',
-            },
-        depends=['process_method'])
+            })
 
     @classmethod
     def __setup__(cls):
@@ -58,6 +66,12 @@ class Journal(metaclass=PoolMeta):
 class Group(metaclass=PoolMeta):
     __name__ = 'account.payment.group'
 
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        cls._buttons['succeed']['invisible'] |= (
+            Eval('process_method') == 'stripe')
+
     def process_stripe(self):
         pool = Pool()
         Payment = pool.get('account.payment')
@@ -78,7 +92,9 @@ class Group(metaclass=PoolMeta):
 
 
 class CheckoutMixin:
-    stripe_checkout_id = fields.Char("Stripe Checkout ID", readonly=True)
+    __slots__ = ()
+    stripe_checkout_id = fields.Char(
+        "Stripe Checkout ID", readonly=True, strip=False)
 
     @classmethod
     def copy(cls, records, default=None):
@@ -110,38 +126,36 @@ class CheckoutMixin:
             }
 
 
-class Payment(CheckoutMixin, metaclass=PoolMeta):
+class Payment(StripeCustomerMethodMixin, CheckoutMixin, metaclass=PoolMeta):
     __name__ = 'account.payment'
 
     stripe_checkout_needed = fields.Function(
         fields.Boolean("Stripe Checkout Needed"),
         'on_change_with_stripe_checkout_needed')
     stripe_charge_id = fields.Char(
-        "Stripe Charge ID", readonly=True,
+        "Stripe Charge ID", readonly=True, strip=False,
         states={
             'invisible': ((Eval('process_method') != 'stripe')
                 | ~Eval('stripe_charge_id')),
-            },
-        depends=['process_method'])
+            })
     stripe_capture = fields.Boolean(
         "Stripe Capture",
         states={
             'invisible': Eval('process_method') != 'stripe',
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['process_method', 'state'])
+            })
     stripe_captured = fields.Boolean(
         "Stripe Captured", readonly=True)
     stripe_capture_needed = fields.Function(
         fields.Boolean("Stripe Capture Needed"),
         'get_stripe_capture_needed')
     stripe_token = fields.Char(
-        "Stripe Token", readonly=True,
+        "Stripe Token", readonly=True, strip=False,
         states={
             'invisible': ~Eval('stripe_token'),
             })
     stripe_payment_intent_id = fields.Char(
-        "Stripe Payment Intent", readonly=True,
+        "Stripe Payment Intent", readonly=True, strip=False,
         states={
             'invisible': ~Eval('stripe_payment_intent_id'),
             })
@@ -150,20 +164,16 @@ class Payment(CheckoutMixin, metaclass=PoolMeta):
         states={
             'invisible': ((Eval('process_method') != 'stripe')
                 | ~Eval('stripe_token')),
-            },
-        depends=['process_method', 'stripe_token'])
+            })
     stripe_capturable = fields.Boolean(
         "Stripe Capturable",
         states={
             'invisible': ((Eval('process_method') != 'stripe')
                 | ~Eval('stripe_payment_intent_id')
                 | ~Eval('stripe_capture_needed')),
-            },
-        depends=[
-            'process_method', 'stripe_payment_intent_id',
-            'stripe_capture_needed'])
+            })
     stripe_idempotency_key = fields.Char(
-        "Stripe Idempotency Key", readonly=True)
+        "Stripe Idempotency Key", readonly=True, strip=False)
     stripe_error_message = fields.Char("Stripe Error Message", readonly=True,
         states={
             'invisible': ~Eval('stripe_error_message'),
@@ -184,76 +194,6 @@ class Payment(CheckoutMixin, metaclass=PoolMeta):
         states={
             'invisible': ~Eval('stripe_dispute_status'),
             })
-    stripe_customer = fields.Many2One(
-        'account.payment.stripe.customer', "Stripe Customer",
-        domain=[
-            ('party', '=', Eval('party', -1)),
-            ('stripe_account', '=', Eval('stripe_account', -1)),
-            ],
-        states={
-            'invisible': Eval('process_method') != 'stripe',
-            'required': Bool(Eval('stripe_customer_source')),
-            'readonly': (~Eval('state').in_(['draft', 'approved'])
-                | Eval('stripe_token') | Eval('stripe_payment_intent_id')),
-            },
-        depends=['party', 'stripe_account', 'process_method',
-            'stripe_customer_source', 'stripe_token',
-            'stripe_payment_intent_id', 'state'])
-    stripe_customer_source = fields.Char(
-        "Stripe Customer Source",
-        states={
-            'invisible': ((Eval('process_method') != 'stripe')
-                | Eval('stripe_token')
-                | Eval('stripe_payment_intent_id')
-                | ~Eval('stripe_customer')),
-            'readonly': ~Eval('state').in_(['draft', 'approved']),
-            },
-        depends=['process_method', 'stripe_token', 'stripe_payment_intent_id',
-            'stripe_customer', 'state'])
-    # Use Function field with selection to avoid to query Stripe
-    # to validate the value
-    stripe_customer_source_selection = fields.Function(fields.Selection(
-            'get_stripe_customer_sources', "Stripe Customer Source",
-            states={
-                'invisible': ((Eval('process_method') != 'stripe')
-                    | Eval('stripe_token')
-                    | Eval('stripe_payment_intent_id')
-                    | ~Eval('stripe_customer')),
-                'readonly': ~Eval('state').in_(['draft', 'approved']),
-                },
-            depends=[
-                'process_method', 'stripe_token', 'stripe_customer', 'state']),
-        'get_stripe_customer_source', setter='set_stripe_customer_source')
-    stripe_customer_payment_method = fields.Char(
-        "Stripe Payment Method",
-        states={
-            'invisible': ((Eval('process_method') != 'stripe')
-                | Eval('stripe_token')
-                | ~Eval('stripe_customer')),
-            'readonly': (~Eval('state').in_(['draft', 'approved'])
-                | Eval('stripe_payment_intent_id')),
-            },
-        depends=['process_method', 'stripe_token', 'stripe_customer', 'state'])
-    # Use Function field with selection to avoid to query Stripe
-    # to validate the value
-    stripe_customer_payment_method_selection = fields.Function(
-        fields.Selection(
-            'get_stripe_customer_payment_methods',
-            "Stripe Customer Payment Method",
-            states={
-                'invisible': ((Eval('process_method') != 'stripe')
-                    | Eval('stripe_token')
-                    | ~Eval('stripe_customer')),
-                'readonly': (~Eval('state').in_(['draft', 'approved'])
-                    | Eval('stripe_payment_intent_id')),
-                },
-            depends=[
-                'process_method', 'stripe_token', 'stripe_customer', 'state']),
-        'get_stripe_customer_payment_method',
-        setter='set_stripe_customer_payment_method')
-    stripe_account = fields.Function(fields.Many2One(
-            'account.payment.stripe.account', "Stripe Account"),
-        'on_change_with_stripe_account')
     stripe_amount = fields.Function(
         fields.Integer("Stripe Amount"),
         'get_stripe_amount', setter='set_stripe_amount')
@@ -263,21 +203,45 @@ class Payment(CheckoutMixin, metaclass=PoolMeta):
             'invisible': ((Eval('process_method') != 'stripe')
                 | (~Eval('stripe_charge_id')
                     & ~Eval('stripe_payment_intent_id'))),
-            },
-        depends=[
-            'process_method', 'stripe_charge_id', 'stripe_payment_intent_id'])
+            })
 
     @classmethod
     def __setup__(cls):
         super(Payment, cls).__setup__()
+
+        cls.stripe_customer.states['readonly'] = (
+            ~Eval('state').in_(['draft', 'submitted', 'approved'])
+            | Eval('stripe_token')
+            | Eval('stripe_payment_intent_id'))
+
+        cls.stripe_customer_source.states['invisible'] |= (
+            Eval('stripe_token') | Eval('stripe_payment_intent_id'))
+        cls.stripe_customer_source.states['readonly'] = (
+            ~Eval('state').in_(['draft', 'submitted', 'approved']))
+
+        cls.stripe_customer_source_selection.states['invisible'] |= (
+            Eval('stripe_token') | Eval('stripe_payment_intent_id'))
+        cls.stripe_customer_source_selection.states['readonly'] = (
+            ~Eval('state').in_(['draft', 'submitted', 'approved']))
+
+        cls.stripe_customer_payment_method.states['invisible'] |= (
+            Eval('stripe_token'))
+        cls.stripe_customer_payment_method.states['readonly'] = (
+            ~Eval('state').in_(['draft', 'submitted', 'approved'])
+            | Eval('stripe_payment_intent_id'))
+
+        cls.stripe_customer_payment_method_selection.states['invisible'] |= (
+            Eval('stripe_token'))
+        cls.stripe_customer_payment_method_selection.states['readonly'] = (
+            ~Eval('state').in_(['draft', 'submitted', 'approved'])
+            | Eval('stripe_payment_intent_id'))
+
         cls.amount.states['readonly'] &= ~Eval('stripe_capture_needed')
-        cls.amount.depends.append('stripe_capture_needed')
         cls.stripe_amount.states.update(cls.amount.states)
-        cls.stripe_amount.depends.extend(cls.amount.depends)
         cls._buttons.update({
                 'stripe_checkout': {
                     'invisible': (~Eval('state', 'draft').in_(
-                            ['approved', 'processing'])
+                            ['submitted', 'approved', 'processing'])
                         | ~Eval('stripe_checkout_needed', False)),
                     'depends': ['state', 'stripe_checkout_needed'],
                     },
@@ -288,20 +252,6 @@ class Payment(CheckoutMixin, metaclass=PoolMeta):
                     },
                 })
 
-    @classmethod
-    def __register__(cls, module_name):
-        cursor = Transaction().connection.cursor()
-        sql_table = cls.__table__()
-        table = cls.__table_handler__(module_name)
-        idempotency_key_exist = table.column_exist('stripe_idempotency_key')
-
-        super(Payment, cls).__register__(module_name)
-
-        # Migration from 4.6: do not set the same key to all existing payments
-        if not idempotency_key_exist:
-            cursor.execute(*sql_table.update(
-                    [sql_table.stripe_idempotency_key], [None]))
-
     @classmethod
     def default_stripe_capture(cls):
         return True
@@ -322,63 +272,6 @@ class Payment(CheckoutMixin, metaclass=PoolMeta):
     def default_stripe_idempotency_key(cls):
         return uuid.uuid4().hex
 
-    @fields.depends('party')
-    def on_change_party(self):
-        super(Payment, self).on_change_party()
-        self.stripe_customer = None
-        self.stripe_customer_source = None
-        self.stripe_customer_source_selection = None
-
-    @fields.depends('stripe_customer', 'stripe_customer_source')
-    def get_stripe_customer_sources(self):
-        sources = [('', '')]
-        if self.stripe_customer:
-            sources.extend(self.stripe_customer.sources())
-        if (self.stripe_customer_source
-                and self.stripe_customer_source not in dict(sources)):
-            sources.append(
-                (self.stripe_customer_source, self.stripe_customer_source))
-        return sources
-
-    @fields.depends(
-        'stripe_customer_source_selection',
-        'stripe_customer_source')
-    def on_change_stripe_customer_source_selection(self):
-        self.stripe_customer_source = self.stripe_customer_source_selection
-
-    def get_stripe_customer_source(self, name):
-        return self.stripe_customer_source
-
-    @classmethod
-    def set_stripe_customer_source(cls, payments, name, value):
-        pass
-
-    @fields.depends('stripe_customer', 'stripe_customer_payment_method')
-    def get_stripe_customer_payment_methods(self):
-        methods = [('', '')]
-        if self.stripe_customer:
-            methods.extend(self.stripe_customer.payment_methods())
-        if (self.stripe_customer_payment_method
-                and self.stripe_customer_payment_method not in dict(methods)):
-            methods.append(
-                (self.stripe_customer_payment_method,
-                    self.stripe_customer_payment_method))
-        return methods
-
-    @fields.depends(
-        'stripe_customer_payment_method_selection',
-        'stripe_customer_payment_method')
-    def on_change_stripe_customer_payment_method_selection(self):
-        self.stripe_customer_payment_method = (
-            self.stripe_customer_payment_method_selection)
-
-    def get_stripe_customer_payment_method(self, name):
-        return self.stripe_customer_payment_method
-
-    @classmethod
-    def set_stripe_customer_payment_method(cls, payments, name, value):
-        pass
-
     @fields.depends('process_method',
         'stripe_token', 'stripe_payment_intent_id',
         'stripe_customer_source', 'stripe_customer_source_selection',
@@ -398,23 +291,17 @@ class Payment(CheckoutMixin, metaclass=PoolMeta):
             and not self.stripe_capture
             and not self.stripe_captured)
 
-    @fields.depends('journal')
-    def on_change_with_stripe_account(self, name=None):
-        if self.journal and self.journal.process_method == 'stripe':
-            return self.journal.stripe_account.id
-
     def get_stripe_amount(self, name):
-        return int(self.amount * 10 ** self.currency_digits)
+        return int(self.amount * 10 ** self.currency.digits)
 
     @classmethod
     def set_stripe_amount(cls, payments, name, value):
-        keyfunc = attrgetter('currency_digits')
+        keyfunc = attrgetter('currency')
         payments = sorted(payments, key=keyfunc)
         value = Decimal(value)
-        for digits, payments in groupby(payments, keyfunc):
-            digits = Decimal(digits)
+        for currency, payments in groupby(payments, keyfunc):
             cls.write(list(payments), {
-                    'amount': value * 10 ** -digits,
+                    'amount': value * 10 ** -Decimal(currency.digits),
                     })
 
     @classmethod
@@ -426,18 +313,21 @@ class Payment(CheckoutMixin, metaclass=PoolMeta):
             ]
 
     @classmethod
-    def validate(cls, payments):
-        super(Payment, cls).validate(payments)
-        for payment in payments:
-            payment.check_stripe_journal()
+    def validate_fields(cls, payments, field_names):
+        super().validate_fields(payments, field_names)
+        cls.check_stripe_journal(payments, field_names)
 
-    def check_stripe_journal(self):
-        if (self.kind != 'receivable'
-                and self.journal.process_method == 'stripe'):
-            raise PaymentValidationError(
-                gettext('account_payment_stripe.msg_stripe_receivable',
-                    journal=self.journal.rec_name,
-                    payment=self.rec_name))
+    @classmethod
+    def check_stripe_journal(cls, payments, field_names=None):
+        if field_names and not (field_names & {'kind', 'journal'}):
+            return
+        for payment in payments:
+            if (payment.kind != 'receivable'
+                    and payment.journal.process_method == 'stripe'):
+                raise PaymentValidationError(
+                    gettext('account_payment_stripe.msg_stripe_receivable',
+                        journal=payment.journal.rec_name,
+                        payment=payment.rec_name))
 
     @classmethod
     def create(cls, vlist):
@@ -499,7 +389,7 @@ class Payment(CheckoutMixin, metaclass=PoolMeta):
         languages = [self.party.lang or Language.get()]
         msg, title = get_email(
             'account.payment.stripe.email_checkout', self, languages)
-        msg['From'] = from_ or from_cfg
+        set_from_header(msg, from_cfg, from_ or from_cfg)
         msg['To'] = ','.join(emails)
         msg['Subject'] = Header(title, 'utf-8')
         sendmail_transactional(from_cfg, emails, msg)
@@ -538,6 +428,8 @@ class Payment(CheckoutMixin, metaclass=PoolMeta):
 
         The transaction is committed after each payment charge.
         """
+        pool = Pool()
+        Customer = pool.get('account.payment.stripe.customer')
         if payments is None:
             payments = cls.search([
                     ('state', '=', 'processing'),
@@ -610,6 +502,10 @@ class Payment(CheckoutMixin, metaclass=PoolMeta):
                 continue
             Transaction().commit()
 
+        customers = [p.stripe_customer for p in payments if p.stripe_customer]
+        if customers:
+            Customer.__queue__.find_identical(customers)
+
     def _charge_parameters(self):
         source, customer = None, None
         if self.stripe_token:
@@ -761,8 +657,7 @@ class Payment(CheckoutMixin, metaclass=PoolMeta):
         "Update payments with Stripe charges"
         for payment in payments:
             charge = payment.stripe_charge_
-            payment_intent = payment.stripe_payment_intent
-            if payment_intent:
+            if payment_intent := payment.stripe_payment_intent:
                 charge = payment_intent.latest_charge
             if charge:
                 payment.stripe_update(charge)
@@ -802,14 +697,12 @@ class Refund(Workflow, ModelSQL, ModelView):
             ],
         states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
-    amount = fields.Numeric(
-        "Amount", required=True, digits=(16, Eval('currency_digits', 2)),
+            })
+    amount = Monetary(
+        "Amount", currency='currency', digits='currency', required=True,
         states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['currency_digits', 'state'])
+            })
     stripe_amount = fields.Function(
         fields.Integer("Stripe Amount"), 'get_stripe_amount')
     reason = fields.Selection([
@@ -820,19 +713,27 @@ class Refund(Workflow, ModelSQL, ModelView):
             ], "Reason",
         states={
             'readonly': Eval('state') != 'draft',
-            },
-        depends=['state'])
+            })
+    submitted_by = employee_field(
+        "Submitted by",
+        states=['submitted', 'approved', 'processing', 'succeeded', 'failed'])
+    approved_by = employee_field(
+        "Approved by",
+        states=['approved', 'processing', 'succeeded', 'failed'])
+
     state = fields.Selection([
             ('draft', "Draft"),
+            ('submitted', "Submitted"),
             ('approved', "Approved"),
             ('processing', "Processing"),
             ('succeeded', "Succeeded"),
             ('failed', "Failed"),
-            ], "State", readonly=True, select=True)
+            ], "State", readonly=True, sort=False)
 
     stripe_idempotency_key = fields.Char(
-        "Stripe Idempotency Key", readonly=True)
-    stripe_refund_id = fields.Char("Stripe Refund ID", readonly=True)
+        "Stripe Idempotency Key", readonly=True, strip=False)
+    stripe_refund_id = fields.Char(
+        "Stripe Refund ID", readonly=True, strip=False)
     stripe_error_message = fields.Char("Stripe Error Message", readonly=True,
         states={
             'invisible': ~Eval('stripe_error_message'),
@@ -846,15 +747,27 @@ class Refund(Workflow, ModelSQL, ModelView):
             'invisible': ~Eval('stripe_error_param'),
             })
 
-    currency_digits = fields.Function(
-        fields.Integer("Currency Digits"), 'on_change_with_currency_digits')
+    currency = fields.Function(
+        fields.Many2One('currency.currency', "Currency"),
+        'on_change_with_currency')
+    company = fields.Function(
+        fields.Many2One('company.company', "Company"),
+        'on_change_with_company')
 
     @classmethod
     def __setup__(cls):
         super().__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.add(
+            Index(
+                t,
+                (t.state, Index.Equality()),
+                where=t.state.in_([
+                        'draft', 'submitted', 'approved', 'processing'])))
         cls.__access__.add('payment')
         cls._transitions |= set((
-                ('draft', 'approved'),
+                ('draft', 'submitted'),
+                ('submitted', 'approved'),
                 ('approved', 'processing'),
                 ('processing', 'succeeded'),
                 ('processing', 'failed'),
@@ -863,10 +776,15 @@ class Refund(Workflow, ModelSQL, ModelView):
                 ))
         cls._buttons.update({
                 'draft': {
-                    'invisible': Eval('state') != 'approved',
+                    'invisible': ~Eval('state').in_(['approved', 'submitted']),
                     'icon': 'tryton-back',
                     'depends': ['state'],
                     },
+                'submit': {
+                    'invisible': Eval('state') != 'draft',
+                    'icon': 'tryton-forward',
+                    'depends': ['state'],
+                    },
                 'approve': {
                     'invisible': Eval('state') != 'draft',
                     'icon': 'tryton-forward',
@@ -875,16 +793,19 @@ class Refund(Workflow, ModelSQL, ModelView):
                 })
 
     def get_stripe_amount(self, name):
-        return int(self.amount * 10 ** self.currency_digits)
+        return int(self.amount * 10 ** self.currency.digits)
 
     @classmethod
     def default_stripe_idempotency_key(cls):
         return uuid.uuid4().hex
 
-    @fields.depends('payment', '_parent_payment.currency_digits')
-    def on_change_with_currency_digits(self, name=None):
-        if self.payment:
-            return self.payment.currency_digits
+    @fields.depends('payment', '_parent_payment.currency')
+    def on_change_with_currency(self, name=None):
+        return self.payment.currency if self.payment else None
+
+    @fields.depends('payment', '_parent_payment.company')
+    def on_change_with_company(self, name=None):
+        return self.payment.company if self.payment else None
 
     @classmethod
     def default_state(cls):
@@ -915,12 +836,21 @@ class Refund(Workflow, ModelSQL, ModelView):
     @classmethod
     @ModelView.button
     @Workflow.transition('draft')
+    @reset_employee('submitted_by', 'approved_by')
     def draft(cls, refunds):
         pass
 
+    @classmethod
+    @ModelView.button
+    @Workflow.transition('submitted')
+    @set_employee('submitted_by')
+    def submit(cls, refunds):
+        pass
+
     @classmethod
     @ModelView.button
     @Workflow.transition('approved')
+    @set_employee('approved_by')
     def approve(cls, refunds):
         pass
 
@@ -1010,8 +940,9 @@ class Account(ModelSQL, ModelView):
     __name__ = 'account.payment.stripe.account'
 
     name = fields.Char("Name", required=True)
-    secret_key = fields.Char("Secret Key", required=True)
-    publishable_key = fields.Char("Publishable Key", required=True)
+    secret_key = fields.Char("Secret Key", required=True, strip=False)
+    publishable_key = fields.Char(
+        "Publishable Key", required=True, strip=False)
     webhook_identifier = fields.Char("Webhook Identifier", readonly=True)
     webhook_endpoint = fields.Function(
         fields.Char(
@@ -1019,13 +950,18 @@ class Account(ModelSQL, ModelView):
             help="The URL to be called by Stripe."),
         'on_change_with_webhook_endpoint')
     webhook_signing_secret = fields.Char(
-        "Webhook Signing Secret",
+        "Webhook Signing Secret", strip=False,
         states={
             'invisible': ~Eval('webhook_identifier'),
             },
-        depends=['webhook_identifier'],
         help="The Stripe's signing secret of the webhook.")
-    last_event = fields.Char("Last Event", readonly=True)
+    last_event = fields.Char("Last Event", readonly=True, strip=False)
+    setup_intent_delay = fields.TimeDelta(
+        "Setup Intent Delay", required=True,
+        domain=[
+            ('setup_intent_delay', '>=', TimeDelta()),
+            ],
+        help="The delay before cancelling setup intent not succeeded.")
 
     @classmethod
     def __setup__(cls):
@@ -1054,6 +990,10 @@ class Account(ModelSQL, ModelView):
                 '/webhook/%(identifier)s'
                 % url_part))
 
+    @classmethod
+    def default_setup_intent_delay(cls):
+        return dt.timedelta(days=30)
+
     @classmethod
     def fetch_events(cls):
         """Fetch last events of each account without webhook and process them
@@ -1429,29 +1369,26 @@ class Customer(CheckoutMixin, DeactivableMixin, ModelSQL, ModelView):
     "Stripe Customer"
     __name__ = 'account.payment.stripe.customer'
     _history = True
-    party = fields.Many2One('party.party', "Party", required=True, select=True,
+    party = fields.Many2One('party.party', "Party", required=True,
         states={
             'readonly': Eval('stripe_customer_id') | Eval('stripe_token'),
-            },
-        depends=['stripe_customer_id', 'stripe_token'])
+            })
     stripe_account = fields.Many2One(
         'account.payment.stripe.account', "Account", required=True,
         states={
             'readonly': Eval('stripe_customer_id') | Eval('stripe_token'),
-            },
-        depends=['stripe_customer_id', 'stripe_token'])
+            })
     stripe_checkout_needed = fields.Function(
         fields.Boolean("Stripe Checkout Needed"), 'get_stripe_checkout_needed')
     stripe_customer_id = fields.Char(
-        "Stripe Customer ID",
+        "Stripe Customer ID", strip=False,
         states={
             'readonly': ((Eval('stripe_customer_id') | Eval('stripe_token'))
                 & (Eval('id', -1) >= 0)),
-            },
-        depends=['stripe_token'])
-    stripe_token = fields.Char("Stripe Token", readonly=True)
+            })
+    stripe_token = fields.Char("Stripe Token", readonly=True, strip=False)
     stripe_setup_intent_id = fields.Char(
-        "Stripe SetupIntent ID", readonly=True)
+        "Stripe SetupIntent ID", readonly=True, strip=False)
     stripe_error_message = fields.Char("Stripe Error Message", readonly=True,
         states={
             'invisible': ~Eval('stripe_error_message'),
@@ -1465,6 +1402,16 @@ class Customer(CheckoutMixin, DeactivableMixin, ModelSQL, ModelView):
             'invisible': ~Eval('stripe_error_param'),
             })
 
+    identical_customers = fields.Many2Many(
+        'account.payment.stripe.customer.identical',
+        'source', 'target', "Identical Customers", readonly=True,
+        states={
+            'invisible': ~Eval('identical_customers'),
+            })
+    fingerprints = fields.One2Many(
+        'account.payment.stripe.customer.fingerprint', 'customer',
+        "Fingerprints", readonly=True)
+
     _sources_cache = Cache(
         'account_payment_stripe_customer.sources',
         duration=config.getint(
@@ -1488,6 +1435,10 @@ class Customer(CheckoutMixin, DeactivableMixin, ModelSQL, ModelView):
                     'invisible': ~Eval('stripe_customer_id'),
                     'depends': ['stripe_customer_id'],
                     },
+                'find_identical': {
+                    'invisible': ~Eval('stripe_customer_id'),
+                    'depends': ['stripe_customer_id'],
+                    },
                 })
 
     def get_stripe_checkout_needed(self, name):
@@ -1581,6 +1532,7 @@ class Customer(CheckoutMixin, DeactivableMixin, ModelSQL, ModelView):
                 # TODO add card
             customer.save()
             Transaction().commit()
+        cls.__queue__.find_identical(customers)
 
     def _customer_parameters(self):
         locales = [pl.lang.code for pl in self.party.langs if pl.lang]
@@ -1791,23 +1743,31 @@ class Customer(CheckoutMixin, DeactivableMixin, ModelSQL, ModelView):
             # Use clear cache after commit
             customer = cls(customer.id)
             setup_intent = customer.stripe_setup_intent
-            if not setup_intent or setup_intent.status != 'succeeded':
+            if not setup_intent:
+                continue
+            if setup_intent.status not in {'succeeded', 'canceled'}:
+                delay = customer.stripe_account.setup_intent_delay
+                expiration = dt.datetime.now() - delay
+                created = dt.datetime.fromtimestamp(setup_intent.created)
+                if created < expiration:
+                    setup_intent.cancel()
                 continue
             customer.lock()
             try:
-                if customer.stripe_customer_id:
-                    stripe.PaymentMethod.attach(
-                        setup_intent.payment_method,
-                        customer=customer.stripe_customer_id,
-                        api_key=customer.stripe_account.secret_key,
-                        stripe_version=STRIPE_VERSION)
-                else:
-                    cu = stripe.Customer.create(
-                        api_key=customer.stripe_account.secret_key,
-                        stripe_version=STRIPE_VERSION,
-                        payment_method=setup_intent.payment_method,
-                        **customer._customer_parameters())
-                    customer.stripe_customer_id = cu.id
+                if setup_intent.status == 'succeeded':
+                    if customer.stripe_customer_id:
+                        stripe.PaymentMethod.attach(
+                            setup_intent.payment_method,
+                            customer=customer.stripe_customer_id,
+                            api_key=customer.stripe_account.secret_key,
+                            stripe_version=STRIPE_VERSION)
+                    else:
+                        cu = stripe.Customer.create(
+                            api_key=customer.stripe_account.secret_key,
+                            stripe_version=STRIPE_VERSION,
+                            payment_method=setup_intent.payment_method,
+                            **customer._customer_parameters())
+                        customer.stripe_customer_id = cu.id
             except (stripe.error.RateLimitError,
                     stripe.error.APIConnectionError) as e:
                 logger.warning(str(e))
@@ -1831,6 +1791,85 @@ class Customer(CheckoutMixin, DeactivableMixin, ModelSQL, ModelView):
             cls._payment_methods_cache.clear()
             Transaction().commit()
 
+    def fetch_fingeprints(self):
+        customer = self.retrieve(expand=['sources'])
+        if customer:
+            for source in customer.sources:
+                if hasattr(source, 'fingerprint'):
+                    yield source.fingerprint
+            try:
+                payment_methods = stripe.PaymentMethod.list(
+                    api_key=self.stripe_account.secret_key,
+                    stripe_version=STRIPE_VERSION,
+                    customer=customer.id,
+                    type='card')
+            except (stripe.error.RateLimitError,
+                    stripe.error.APIConnectionError) as e:
+                logger.warning(str(e))
+                payment_methods = []
+            for payment_method in payment_methods:
+                yield payment_method.card.fingerprint
+
+    @classmethod
+    @ModelView.button
+    def find_identical(cls, customers):
+        pool = Pool()
+        Fingerprint = pool.get('account.payment.stripe.customer.fingerprint')
+        new = []
+        for customer in customers:
+            fingerprints = set(customer.fetch_fingeprints())
+            fingerprints -= {f.fingerprint for f in customer.fingerprints}
+            for fingerprint in fingerprints:
+                new.append(Fingerprint(
+                        customer=customer,
+                        fingerprint=fingerprint))
+        Fingerprint.save(new)
+
+
+class CustomerFingerprint(ModelSQL):
+    "Stripe Customer Fingerprint"
+    __name__ = 'account.payment.stripe.customer.fingerprint'
+    customer = fields.Many2One(
+        'account.payment.stripe.customer', "Customer",
+        required=True, ondelete='CASCADE')
+    fingerprint = fields.Char("Fingerprint", required=True)
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        t = cls.__table__()
+        cls._sql_constraints += [
+            ('customer_fingerprint_unique',
+                Unique(t, t.customer, t.fingerprint),
+                'account_payment_stripe.msg_customer_fingerprint_unique'),
+            ]
+
+
+class CustomerIdentical(ModelSQL):
+    "Stripe Customer Identical"
+    __name__ = 'account.payment.stripe.customer.identical'
+    source = fields.Many2One('account.payment.stripe.customer', "Source")
+    target = fields.Many2One('account.payment.stripe.customer', "Target")
+
+    @classmethod
+    def table_query(cls):
+        pool = Pool()
+        Fingerprint = pool.get('account.payment.stripe.customer.fingerprint')
+        source = Fingerprint.__table__()
+        target = Fingerprint.__table__()
+        return (
+            source
+            .join(target, condition=source.fingerprint == target.fingerprint)
+            .select(
+                Literal(0).as_('create_uid'),
+                source.create_date.as_('create_date'),
+                Literal(None).as_('write_uid'),
+                Literal(None).as_('write_date'),
+                sql_pairing(source.id, target.id).as_('id'),
+                source.customer.as_('source'),
+                target.customer.as_('target'),
+                where=source.customer != target.customer))
+
 
 class Checkout(Wizard):
     "Stripe Checkout"
